import numpy as np 
import struct
import polyscope as ps
from scipy.spatial import ckdtree
#####
#####
#I/O TOOLS
#####
#####

def separate_faces_dict(Faces,n_towers=10): 
    n_towers = np.amax(Faces[:,[3,4]])+1
   
    Occupancy=np.zeros(n_towers)
    Dict={}
    for face in Faces : 
        _,_,_,num1,num2 = face
        if num1!=-1:
            if Occupancy[num1]==0:
                Dict[num1]=[face[[0,1,2]]]
                Occupancy[num1]+=1
            else : 
                Dict[num1].append(face[[0,1,2]])
            
        if num2!=-1:
            if Occupancy[num2]==0:
                Dict[num2]=[face[[0,1,2]]]
                Occupancy[num2]+=1
            else : 
                Dict[num2].append(face[[0,1,2]])
            
            
    Faces_separated={}
    for i in sorted(Dict.keys()) : 
        Faces_separated[i] = (np.array(Dict[i]))
        
    return(Faces_separated)

def write_mesh_bin(filename, Verts, Faces):
    assert(len(Faces[0])==5 and len(Verts[0])==3)
    strfile = struct.pack("Q", len(Verts))
    strfile +=Verts.flatten().astype(np.float64).tobytes()
    strfile += struct.pack("Q", len(Faces))
    dt=np.dtype([('triangles',np.uint64,(3,)), ('labels',np.int32,(2,))])
    F_n = Faces[:,:3].astype(np.uint64)
    F_t = Faces[:,3:].astype(np.int32)
    func = lambda i : (F_n[i],F_t[i])
    T=np.array(list(map(func,np.arange(len(Faces)))),dtype=dt)
    strfile+=T.tobytes()
    file = open(filename,'wb')
    file.write(strfile)
    file.close()


def write_mesh_text(filename, Verts, Faces):
    
    file = open(filename, 'w')
    file.write(str(len(Verts))+'\n')
    for i in range(len(Verts)): 
        file.write(f'{Verts[i][0]:.5f} {Verts[i][1]:.5f} {Verts[i][2]:.5f}'+'\n')
    file.write(str(len(Faces))+'\n')
    for i in range(len(Faces)): 
        file.write(f'{Faces[i][0]} {Faces[i][1]} {Faces[i][2]} {Faces[i][3]} {Faces[i][4]}'+'\n')
    file.close() 

def open_mesh_multitracker(filename):
    try : 
        return(read_rec_file_bin(filename))
    except : 
        return(read_rec_file_num(filename))

def read_rec_file_bin(filename): 
    mesh_file=open(filename,'rb')
    
    #Vertices
    num_vertices,=struct.unpack('Q', mesh_file.read(8)) 
    Verts=np.fromfile(mesh_file,count=3*num_vertices,dtype=np.float64).reshape((num_vertices,3))

    #Triangles
    num_triangles,=struct.unpack('Q', mesh_file.read(8))

    # dtype # 3 unsigned integers (long long) for the triangles # 2 integers for the labels
    dt=np.dtype([('triangles',np.uint64,(3,)), ('labels',np.int32,(2,))])
    t=np.fromfile(mesh_file,count=num_triangles,dtype=dt)
    mesh_file.close()

    Faces_num=t['triangles']
    Faces_labels=t['labels'] - 1
    Faces = np.hstack((Faces_num,Faces_labels))
    return(Verts, Faces.astype(int),np.array([num_vertices,num_triangles]))

def read_rec_file_num(filename,offset=-1): 
    mesh_file = open(filename, 'rb')
    Ns= []
    Verts = []
    Faces= []

    Lines=[]
    for line in mesh_file.readlines():
        L=line.decode('UTF8')
        L=L[:-1].split(' ')
        Lines.append(L)
        if len(L)==1 : 
            Ns.append(L[0])
        elif len(L)==3 : 
            Verts.append(L)
        else : 
            Faces.append(L)
    mesh_file.close()

    Faces = np.array(Faces).astype(int)
    Faces[:,[3,4]]+=offset
    Verts = np.array(Verts).astype(float)
    Ns = np.array(Ns).astype(int)
    return(Verts, Faces, Ns)







#####
#####
#Mesh cleaning
#####
#####

def retrieve_mesh_multimaterial_multitracker_format(Graph,Map):
    ##Must be used without any filtering operation
    reverse_map ={}
    for key in Map : 
        for node_idx in Map[key] :
            reverse_map[node_idx]=key
    Faces=[]
    Faces_idx = []
    Nodes_linked = []
    for idx in range(len(Graph.Faces)) : 
        nodes_linked = Graph.Nodes_Linked_by_Faces[idx]
        

        cluster_1 = reverse_map[nodes_linked[0]]
        cluster_2 = reverse_map[nodes_linked[1]]

        if cluster_1 != cluster_2 : 
            
            face = Graph.Faces[idx]
            cells = [cluster_1,cluster_2]
            Faces.append([face[0],face[1],face[2],cells[0],cells[1]])
            Faces_idx.append(idx)
            Nodes_linked.append(nodes_linked)

    for idx in range(len(Graph.Lone_Faces)):
        face = Graph.Lone_Faces[idx]
        node_linked = Graph.Nodes_linked_by_lone_faces[idx]
        cluster_1 = reverse_map[node_linked]
        #We incorporate all these edges because they are border edges
        if cluster_1 !=0:
            cells = [0,cluster_1]
            Faces.append([face[0],face[1],face[2],cells[0],cells[1]])
            Faces_idx.append(idx)
            Nodes_linked.append(nodes_linked)

    return(Graph.Vertices, np.array(Faces),Faces_idx,np.array(Nodes_linked))


def Clean_mesh_from_seg(Seg):
    #Take a Segmentation class as entry
    
    V, Faces, Faces_idx, Nodes_linked = retrieve_mesh_multimaterial_multitracker_format(Seg.Delaunay_Graph,Seg.Map_end)
    Verts = V.copy()
    
    for i, f in enumerate(Faces): 
        if f[3]>f[4]: 
            Faces[i]=Faces[i,[0,1,2,4,3]]
            Nodes_linked[i]=Nodes_linked[i][[1,0]]
            
    
    Faces = reorient_faces(Faces,Seg,Nodes_linked)

    for i in range(len(Faces)): 
        Faces[i]=Faces[i,[0,2,1,3,4]]
            
    return(Verts,Faces)


def compute_normal_Faces(Verts,Faces):
    Pos = Verts[Faces[:,[0,1,2]]]
    Sides_1 = Pos[:,1]-Pos[:,0]
    Sides_2 = Pos[:,2]-Pos[:,1]
    Normal_faces = np.cross(Sides_1,Sides_2,axis=1)
    Norms = np.linalg.norm(Normal_faces,axis=1)#*(1+1e-8)
    Normal_faces/=(np.array([Norms]*3).transpose())
    return(Normal_faces)

def reorient_faces(Faces,Seg,Nodes_linked):
       
    #Thumb rule for all the faces
    
    Normals = compute_normal_Faces(Seg.Delaunay_Graph.Vertices,Faces)
    
    P = Seg.Delaunay_Graph.Vertices[Faces[:,:3]]
    Centroids_faces = np.mean(P,axis=1)
    Centroids_nodes = np.mean(Seg.Delaunay_Graph.Vertices[Seg.Delaunay_Graph.Tetra[Nodes_linked[:,0]]],axis=1)
    Vectors = Centroids_nodes-Centroids_faces
    Norms = np.linalg.norm(Vectors,axis=1)
    Vectors[:,0]/=Norms
    Vectors[:,1]/=Norms
    Vectors[:,2]/=Norms

    Dot_product = np.sum(np.multiply(Vectors,Normals),axis=1)
    Normals_sign = np.sign(Dot_product)
    
    #Reorientation according to the normal sign
    reoriented_faces = Faces.copy()
    for i,s in enumerate(Normals_sign) : 
        if s <0 : 
            reoriented_faces[i]=reoriented_faces[i][[0,2,1,3,4]]
            
    return(reoriented_faces)


#####
#####
#MESH PLOTTING
#####
#####


def retrieve_border_tetra_with_index_map(Graph,Map):
    reverse_map ={}
    for key in Map : 
            for node_idx in Map[key] :
                reverse_map[node_idx]=key
                
    Clusters=[]
    for _ in range(len(Map)): 
        Clusters.append([])

    for idx in range(len(Graph.Faces)) : 
        nodes_linked = Graph.Nodes_Linked_by_Faces[idx]

        cluster_1 = reverse_map.get(nodes_linked[0],-1)
        cluster_2 =reverse_map.get(nodes_linked[1],-2)
        #if the two nodes of the edges belong to the same cluster we ignore them
        #otherwise we add them to the mesh
        if cluster_1 != cluster_2 : 
            face = Graph.Faces[idx]
            if cluster_1 >= 0 : 
                Clusters[cluster_1].append(face)
            if cluster_2 >= 0 : 
                Clusters[cluster_2].append(face)
            

    for idx in range(len(Graph.Lone_Faces)):
        edge = Graph.Lone_Faces[idx]
        node_linked = Graph.Nodes_linked_by_lone_faces[idx]
        cluster_1 = reverse_map[node_linked]
        #We incorporate all these edges because they are border edges
        if cluster_1 !=0:
            v1,v2,v3=edge[0],edge[1],edge[2]
            Clusters[cluster_1].append([v1,v2,v3])
    return(Clusters)


def compute_seeds_idx_from_voxel_coords(EDT,Centroids,Coords):
    ##########
    ## Compute the seeds used for watershed
    ##########
    
    nx,ny,nz = EDT.shape
    Points = create_coords(nx,ny,nz)

    Anchors = Coords[:,0]*ny*nz+Coords[:,1]*nz+Coords[:,2]
    p=Points[Anchors]

    tree = ckdtree.cKDTree(Centroids)
    Dist,Idx_seeds = tree.query(p)
    return(Idx_seeds)
    #unique,indices = np.unique(Idx_seeds,return_index=True)
    #return(Idx_seeds[indices])
    #return(Idx_seeds[sorted(indices)])

def create_coords(nx,ny,nz):
    XV = np.linspace(0,nx-1,nx)
    YV = np.linspace(0,ny-1,ny)
    ZV = np.linspace(0,nz-1,nz)
    xvv, yvv, zvv = np.meshgrid(XV,YV,ZV)
    xvv=np.transpose(xvv,(1,0,2)).flatten()
    yvv=np.transpose(yvv,(1,0,2)).flatten()
    zvv=zvv.flatten()
    Points=np.vstack(([xvv,yvv,zvv])).transpose().astype(int)
    return(Points)


def plot_cells_polyscope(Verts,Faces,clean_before = True, clean_after = True):
    Clusters = separate_faces_dict(Faces)
    
    ps.init()
    
    ps.set_ground_plane_mode("none")
    if clean_before : 
        ps.remove_all_structures()
        
    for key in sorted(Clusters.keys()):
        cluster = Clusters[key]
        ps.register_surface_mesh("Cell "+str(key), Verts, np.array(cluster), smooth_shade=False)
    ps.show()
    
    if clean_after : 
        ps.remove_all_structures()
    

    
def renormalize_verts(Verts,Faces): 
    #When the Vertices are only a subset of the faces, we remove the useless vertices and give the new faces
    idx_Verts_used = np.unique(Faces)
    Verts_used = Verts[idx_Verts_used]
    idx_mapping = np.arange(len(Verts_used))
    mapping = dict(zip(idx_Verts_used,idx_mapping))
    def func(x): 
        return([mapping[x[0]],mapping[x[1]],mapping[x[2]]])
    New_Faces = np.array(list(map(func,Faces)))
    return(Verts_used,New_Faces)






from numpy.linalg import eig
import networkx as nx
from .Tension_inference import infer_forces, infer_tension, infer_pressure, compute_residual_junctions_dict

from dw3d.Dcel import compute_faces_areas,compute_normal_Faces, update_graph_with_scattered_values
from dw3d.Curvature import compute_gaussian_curvature_vertices, compute_curvature_vertices_cotan,compute_sphere_fit_residues_dict

from .Plotting_utilities import *

def plot_force_inference(Mesh,dict_tensions = None, dict_pressure = None,alpha = 0.05, scalar_quantities = False,scattered = False,scatter_coeff=0.2):
    
    if dict_tensions == None : 
        dict_tensions,dict_pressure = infer_forces(Mesh)
    if dict_pressure == None: 
        dict_pressure = infer_pressure(Mesh, dict_tensions)
    
    G=Mesh.compute_networkx_graph()
    nx.set_edge_attributes(G, dict_tensions, "tension")
    nx.set_node_attributes(G, dict_pressure, "pressure")
    
    if scattered : 
        Mesh.compute_scattered_arrays(coeff = scatter_coeff)
        G = update_graph_with_scattered_values(G,Mesh)

    dict_sphere_fit_residual = compute_sphere_fit_residues_dict(G)
    
    ps_mesh = view_dict_values_on_mesh(Mesh,dict_tensions,alpha = alpha,ps_mesh = None, scattered=scattered, clean_before = False, clean_after=False,show=False,name_values = "Surface Tensions")
    ps_mesh = view_pressures_on_mesh(Mesh,dict_pressure,ps_mesh = ps_mesh, alpha = alpha, clean_before=False, clean_after=False, show = False,scattered=scattered)
    plot_stress_tensor(Mesh,G,dict_tensions, dict_pressure, clean_before = False, clean_after = False,show=False)
    display_embryo_graph_forces(G,alpha = alpha, clean_before = False, clean_after = False,show=False,P0 = 0, Plot_pressures = True)
    
    if scalar_quantities : 
        ps_mesh = view_dict_values_on_mesh(Mesh,dict_sphere_fit_residual,alpha = alpha,ps_mesh = ps_mesh,scattered=scattered, clean_before = False, clean_after=False,show=False,name_values = "Sphere fit residual")
        view_area_derivatives(Mesh, alpha = alpha,ps_mesh = ps_mesh, clean_before = False, clean_after = False,show=False,scattered=scattered)
        view_volume_derivatives(Mesh, alpha = alpha,ps_mesh = ps_mesh, clean_before = False, clean_after = False,show=False,scattered=scattered)
        view_mean_curvature_cotan(Mesh, alpha = alpha,ps_mesh = ps_mesh, clean_before = False, clean_after = False,show=False,scattered=scattered)
        view_mean_curvature_robust(Mesh, alpha = alpha,ps_mesh = ps_mesh, clean_before = False, clean_after = False,show=False,scattered=scattered)
        view_gaussian_curvature(Mesh, alpha = alpha,ps_mesh = ps_mesh, clean_before = False, clean_after = False,show=False,scattered=scattered)
        view_discrepancy_of_principal_curvatures(Mesh, alpha = alpha,ps_mesh = ps_mesh, clean_before = False, clean_after = False,show=False,scattered=scattered)

    ps.show()
    

def plot_tension_inference(Mesh,dict_tensions=None,alpha = 0.05, scalar_quantities = False,scattered = False,scatter_coeff=0.2):
    
    ps.remove_all_structures()
    
    if dict_tensions == None : 
        _,dict_tensions,_ = infer_tension(Mesh,mean_tension=1)
        
    G=Mesh.compute_networkx_graph()
    nx.set_edge_attributes(G, dict_tensions, "tension")
    
    if scattered : 
        Mesh.compute_scattered_arrays(coeff = scatter_coeff)
        G = update_graph_with_scattered_values(G,Mesh)
        
    dict_sphere_fit_residual = compute_sphere_fit_residues_dict(G)
    
    
    ps_mesh = view_dict_values_on_mesh(Mesh,dict_tensions,alpha = alpha,ps_mesh = None, scattered=scattered,clean_before = False, clean_after=False,show=False,name_values = "Surface Tensions")

    display_embryo_graph_forces(G,alpha = alpha, clean_before = False, clean_after = False,show=False,P0 = 0, Plot_pressures = False)
    if scalar_quantities : 
        ps_mesh = view_dict_values_on_mesh(Mesh,dict_sphere_fit_residual,alpha = alpha,ps_mesh = ps_mesh,scattered=scattered, clean_before = False, clean_after=False,show=False,name_values = "Sphere fit residual")
        view_area_derivatives(Mesh, alpha = alpha,ps_mesh = ps_mesh, clean_before = False, clean_after = False,show=False,scattered=scattered)
        view_volume_derivatives(Mesh, alpha = alpha,ps_mesh = ps_mesh, clean_before = False, clean_after = False,show=False,scattered=scattered)
        view_mean_curvature_cotan(Mesh, alpha = alpha,ps_mesh = ps_mesh, clean_before = False, clean_after = False,show=False,scattered=scattered)
        view_mean_curvature_robust(Mesh, alpha = alpha,ps_mesh = ps_mesh, clean_before = False, clean_after = False,show=False,scattered=scattered)
        view_gaussian_curvature(Mesh, alpha = alpha,ps_mesh = ps_mesh, clean_before = False, clean_after = False,show=False,scattered=scattered)
        view_discrepancy_of_principal_curvatures(Mesh, alpha = alpha,ps_mesh = ps_mesh, clean_before = False, clean_after = False,show=False,scattered=scattered)

    ps.show()
    

def view_pressures_on_mesh(Mesh,dict_pressures,alpha = 0.05,ps_mesh = None, clean_before = True, clean_after=True,show=True, scattered = False):

    if scattered:
        v,f,cluster_idx = Mesh.v_scattered,Mesh.f_scattered,Mesh.cluster_idx_scattered#, Mesh.idx_scattered
        def find_pressure(idx): 
            return(dict_pressures[idx])
        pressures = np.array(list(map(find_pressure,cluster_idx)))
    else : 
        v,f = Mesh.v,Mesh.f
        def find_pressure(face): 
            return(dict_pressures[max(face[3],face[4])])
        pressures = np.array(list(map(find_pressure,f)))
    
    maxp = np.quantile(pressures,1-alpha)

    print("Extremas of pressures plotted: ",0,maxp)

    pressures = pressures.clip(0,maxp)
    pressures/=np.amax(pressures)
    pressures = 1-pressures
    #ps_mesh = view_faces_values_on_embryo(Mesh,pressures,ps_mesh = ps_mesh,name_values = "Pressures Cells",colormap = cm.magma,clean_before = clean_before, clean_after=clean_after,show=show,adapt_values=False,scattered = False)

    if ps_mesh == None : 
        ps_mesh = ps.register_surface_mesh("Embryo", v,f[:,[0,1,2]])
    ps_mesh.set_color((0.3, 0.6, 0.8)) # rgb triple on [0,1]
    colors_face = cm.magma(pressures)[:,:3]
    ps_mesh.add_color_quantity("Pressures Cells", colors_face, defined_on='faces',enabled=True)
    return(ps_mesh)


"""
def view_pressures_on_mesh(Mesh,dict_pressures,alpha = 0.05,ps_mesh = None, clean_before = True, clean_after=True,show=True, scattered = False):

    v,f = Mesh.v,Mesh.f

    def find_pressure(face): 
        return(dict_pressures[max(face[3],face[4])])

    pressures = np.array(list(map(find_pressure,f)))
    pressures_values = np.array(list(dict_pressures.values()))
    maxp = np.quantile(pressures_values,1-alpha)

    print("Extremas of pressures plotted: ",0,maxp)

    pressures = pressures.clip(0,maxp)
    pressures/=np.amax(pressures)
    pressures = 1-pressures
    ps_mesh = view_faces_values_on_embryo(Mesh,pressures,ps_mesh = ps_mesh,name_values = "Pressures Cells",colormap = cm.magma,clean_before = clean_before, clean_after=clean_after,show=show,adapt_values=False,scattered = scattered)

    return(ps_mesh)



def view_faces_values_on_embryo(Mesh,Vf,name_values = "Values",ps_mesh = None,colormap = cm.jet,min_to_zero=True,clean_before = True, clean_after=True,show=True,highlight_junctions=False,adapt_values = True, scattered = False):
    
    if scattered : 
        v,f,idx = Mesh.v_scattered,Mesh.f_scattered, Mesh.idx_scattered
    
    else : 
        v,f,idx = Mesh.v,Mesh.f, np.arange(len(Mesh.f))
        
    Values = Vf.copy()
    if adapt_values : 
        if min_to_zero : 
            Values-=np.amin(Values)
        Values/=(np.amax(Values)-np.amin(Values))

    Values = Values[idx]
    colors_face = colormap(Values)[:,:3]

    ps.init()
    
    if clean_before : 
        ps.remove_all_structures()


    if ps_mesh == None : 
        ps_mesh = ps.register_surface_mesh("Embryo", v,f[:,[0,1,2]])

    ps_mesh.set_color((0.3, 0.6, 0.8)) # rgb triple on [0,1]
    #ps_mesh.set_transparency(0.2)
    ps_mesh.add_color_quantity(name_values, colors_face, defined_on='faces',enabled=True)
    
    
    if highlight_junctions : 
        plot_trijunctions(Mesh,clean_before = False, clean_after = False, show=False, color = "uniform",value_color = np.ones(3))
    
    ps.set_ground_plane_mode("none") 
    
    if show : 
        ps.show()
    
    if clean_after : 
        ps.remove_all_structures()
        
    return(ps_mesh)

"""
###
#Scalar quantities
###

def view_area_derivatives(Mesh,alpha = 0.05,ps_mesh = None, remove_trijunctions=True,clean_before = True, clean_after = True,show=True,scattered=False):
    DA = Mesh.compute_area_derivatives()
    Derivatives_verts = np.zeros(Mesh.v.shape)
    for key in DA.keys() : 
        Derivatives_verts +=np.abs(DA[key])

    vmin, vmax = np.quantile(Derivatives_verts,alpha),np.quantile(Derivatives_verts,1-alpha)
    print("Extremas of area derivatives plotted: ",vmin,vmax)
    Derivatives_verts = Derivatives_verts.clip(vmin,vmax)
    
    ps_mesh = view_vertex_values_on_embryo(Mesh,Derivatives_verts,name_values="Area Derivatives",ps_mesh = ps_mesh,remove_trijunctions=remove_trijunctions,clean_before = clean_before, clean_after = clean_after,show=show, scattered = scattered)
    
    return(ps_mesh)

def view_volume_derivatives(Mesh,alpha = 0.05,ps_mesh = None, remove_trijunctions=True,clean_before = True, clean_after = True,show=True,scattered=False):
    DV = Mesh.compute_volume_derivatives()
    Derivatives_verts = np.zeros(Mesh.v.shape)
    for dv in DV.keys() : 
        Derivatives_verts +=np.abs(DV[dv])
    
    vmin, vmax = np.quantile(Derivatives_verts,alpha),np.quantile(Derivatives_verts,1-alpha)
    print("Extremas of volume derivatives plotted: ",vmin,vmax)
    Derivatives_verts = Derivatives_verts.clip(vmin,vmax)
    
    ps_mesh = view_vertex_values_on_embryo(Mesh,Derivatives_verts,ps_mesh = ps_mesh, name_values = "Volume Derivatives",remove_trijunctions=remove_trijunctions,clean_before = clean_before, clean_after = clean_after,show=show, scattered = scattered)
    
    return(ps_mesh)




def view_mean_curvature_cotan(Mesh,alpha = 0.05,ps_mesh = None, remove_trijunctions=True,clean_before = True, clean_after = True,show=True,scattered = False):
    
    H,_,_=compute_curvature_vertices_cotan(Mesh)
    
    vmin, vmax = np.quantile(H,alpha),np.quantile(H,1-alpha)
    print("Extremas of mean curvature (cotan) plotted: ",vmin,vmax)
    H = H.clip(vmin,vmax)

    ps_mesh = view_vertex_values_on_embryo(Mesh,H,ps_mesh = ps_mesh,name_values="Mean Curvature Cotan",remove_trijunctions=remove_trijunctions,clean_before = clean_before, clean_after = clean_after,show=show, scattered = scattered)
        
    return(ps_mesh)

def view_gaussian_curvature(Mesh,alpha = 0.05,ps_mesh = None, remove_trijunctions=True,clean_before = True, clean_after = True,show=True,scattered = False):
    
    H = compute_gaussian_curvature_vertices(Mesh)
    
    vmin, vmax = np.quantile(H,alpha),np.quantile(H,1-alpha)
    print("Extremas of gaussian curvature plotted: ",vmin,vmax)
    H = H.clip(vmin,vmax)

    ps_mesh = view_vertex_values_on_embryo(Mesh,H,ps_mesh = ps_mesh,name_values="Gaussian Curvature",remove_trijunctions=remove_trijunctions,clean_before = clean_before, clean_after = clean_after,show=show, scattered = scattered)
        
    return(ps_mesh)



def view_discrepancy_of_principal_curvatures(Mesh,alpha = 0.05,ps_mesh = None, remove_trijunctions=True,clean_before = True, clean_after = True,show=True,scattered = False):
    
    H,_,_=compute_curvature_vertices_cotan(Mesh)
    H/=9
    H[H==0]=np.mean(H)
    
    G = compute_gaussian_curvature_vertices(Mesh)
    Kdiff =  np.abs((H**2-G)/(H**2))

    vmin, vmax = np.quantile(Kdiff,alpha),np.quantile(Kdiff,1-alpha)
    print("Extremas of discrepancy of principal curvatures plotted: ",vmin,vmax)
    Kdiff = Kdiff.clip(vmin,vmax)

    ps_mesh = view_vertex_values_on_embryo(Mesh,Kdiff,ps_mesh = ps_mesh,name_values="Principal curvature discrepancy",remove_trijunctions=remove_trijunctions,clean_before = clean_before, clean_after = clean_after,show=show, scattered = scattered)
        
    return(ps_mesh)










###
#Stress and graph
###


def plot_stress_tensor(Mesh, G, dict_tensions,dict_pressure,clean_before = True, clean_after = True,show=True,lumen_materials=[0]):

    # Formula from G. K. BATCHELOR, J Fluid Mech, 1970
    
    Vols = Mesh.compute_volumes_cells()
    
    Membrane_in_contact_with_cells = {key:[] for key in Mesh.materials}

    for key in dict_tensions.keys() : 
        a,b = key
        Membrane_in_contact_with_cells[a].append(key)
        Membrane_in_contact_with_cells[b].append(key)

    dict_faces_membrane = dict(zip(dict_tensions.keys(),[[] for i in range(len(dict_tensions.keys()))]))
    for i,face in enumerate(Mesh.f) :
        a,b = face[3:]
        dict_faces_membrane[(a,b)].append(i)
        
        
    Areas_triangles = compute_faces_areas(Mesh.v,Mesh.f)
    Normals_triangles = compute_normal_Faces(Mesh.v,Mesh.f)
    delta = np.identity(3)
    tensor_normal_normal = lambda x : delta-np.tensordot(x,x,axes=0)
    Tensordot_normal_triangles = np.array(list(map(tensor_normal_normal,Normals_triangles)))

    for i in range(len(Tensordot_normal_triangles)):
        Tensordot_normal_triangles[i]*=Areas_triangles[i]


    Stress_vectors = np.zeros((Mesh.n_materials-1,3,3))
    Compression_vectors = np.zeros((Mesh.n_materials-1,3,3))
    #compression_dots = []
    delta = np.identity(3)
    for i in range(1,Mesh.n_materials): 
        c = Mesh.materials[i]
        if G.nodes.data('volume')[c]<=0 :
            continue
        mean_stress = np.zeros((3,3))
        p = dict_pressure[c]
        v = Vols[c]
        mean_stress += -p*delta

        for m in Membrane_in_contact_with_cells[c]:
            a,b = m 
            
            #lumen_materials is by default 0
            #A membrane in contact with the exterior medium (i.e in lumen_materials) is counted once. 
            #A membrane in contact with another cell is counted twice, thus we have to divide its surface tension by 2
            #See Supplementary Note for further explanations
            if (a in lumen_materials) or (b in lumen_materials):
                t = dict_tensions[m]
            else : 
                t = dict_tensions[m]/2
                
            Sum=0
            for nt in dict_faces_membrane[m]:
                Sum+=Tensordot_normal_triangles[nt]
            if a == c : #we need to reverse the triangle !
                sign = -1
            elif b ==c : 
                sign = 1

            mean_stress += t/(v) * sign * Sum


        vals, vects = eig(mean_stress)

        Stress_vectors[i-1] = vects.copy()
        Compression_vectors[i-1] = vects.copy()


        if vals[0]<0 : 
            Stress_vectors[i-1,:,0] *= vals[0]
            Compression_vectors[i-1,:,0] *= 0
        else : 
            Stress_vectors[i-1,:,0] *= 0
            Compression_vectors[i-1,:,0]*= vals[0]

        if vals[1]<0 : 
            Stress_vectors[i-1,:,1] *= vals[1]
            Compression_vectors[i-1,:,1] *= 0
        else : 
            Stress_vectors[i-1,:,1] *= 0
            Compression_vectors[i-1,:,2]*= vals[1]

        if vals[2]<0 : 
            Stress_vectors[i-1,:,2] *= vals[2]
            Compression_vectors[i-1,:,2] *= 0
        else : 
            Stress_vectors[i-1,:,2] *= 0
            Compression_vectors[i-1,:,2]*= vals[2]

        #if vals[0]>0 or vals[1]>0 or vals[2]>0 : 
        #    compression_dots.append(G.nodes.data('centroid')[c])
    #compression_dots = np.array(compression_dots)
    
    
    ###
    #PLOTING
    ###
    
    ps.init()
    ps.set_ground_plane_mode("none")
    if clean_before : 
        ps.remove_all_structures()

    # register a point cloud
    N = Mesh.n_materials-1
    centroids = np.array([a[1] for a in G.nodes.data('centroid')])[np.array(G.nodes.data('volume'))[:,1]>0]
    
    #ps_mesh = ps.register_surface_mesh("volume mesh", Mesh.v, Mesh.f[:,[0,1,2]])

    #ps_mesh.set_enabled() # default is true

    #ps_mesh.set_color((0.3, 0.6, 0.8)) # rgb triple on [0,1]
    #ps_mesh.set_transparency(0.2)

    ps_cloud = ps.register_point_cloud("Stress tensors principal directions", centroids,color=(0,0,0),radius=0.004)

    
    
    #For extensile stress:
    vecs_0 = Stress_vectors[:,:,0][np.array(G.nodes.data('volume'))[1:,1]>0]
    vecs_1 = Stress_vectors[:,:,1][np.array(G.nodes.data('volume'))[1:,1]>0]
    vecs_2 = Stress_vectors[:,:,2][np.array(G.nodes.data('volume'))[1:,1]>0]
    radius = 0.005
    length = 0.1
    color = (0.2, 0.5, 0.5)
    # basic visualization
    all_vecs = np.vstack((vecs_0,-vecs_0, vecs_1,-vecs_1, vecs_2, -vecs_2))
    ps_cloud.add_vector_quantity("principal_axes_0", vecs_0, enabled=True,radius=radius, length=length, color=color)
    ps_cloud.add_vector_quantity("principal_axes_0_down", -vecs_0, enabled=True,radius=radius, length=length, color=color)
    ps_cloud.add_vector_quantity("principal_axes_1", vecs_1, enabled=True,radius=radius, length=length, color=color)
    ps_cloud.add_vector_quantity("principal_axes_1_down", -vecs_1, enabled=True,radius=radius, length=length, color=color)
    ps_cloud.add_vector_quantity("principal_axes_2", vecs_2, enabled=True,radius=radius, length=length, color=color)
    ps_cloud.add_vector_quantity("principal_axes_2_down", -vecs_2, enabled=True,radius=radius, length=length, color=color)
    
    #For compressive stress:
    vecs_comp_0 = Compression_vectors[:,:,0][np.array(G.nodes.data('volume'))[1:,1]>0]
    vecs_comp_1 = Compression_vectors[:,:,1][np.array(G.nodes.data('volume'))[1:,1]>0]
    vecs_comp_2 = Compression_vectors[:,:,2][np.array(G.nodes.data('volume'))[1:,1]>0]
    red = (0.7, 0., 0.)
    all_vecs = np.vstack((vecs_0,-vecs_0, vecs_1,-vecs_1, vecs_2, -vecs_2))
    ps_cloud.add_vector_quantity("compressive_stress__axes_0", vecs_comp_0, enabled=True,radius=radius, length=length, color=red)
    ps_cloud.add_vector_quantity("compressive_stress__axes_0_down", -vecs_comp_0, enabled=True,radius=radius, length=length, color=red)
    ps_cloud.add_vector_quantity("compressive_stress__axes_1", vecs_comp_1, enabled=True,radius=radius, length=length, color=red)
    ps_cloud.add_vector_quantity("compressive_stress__axes_1_down", -vecs_comp_1, enabled=True,radius=radius, length=length, color=red)
    ps_cloud.add_vector_quantity("compressive_stress__axes_2", vecs_comp_2, enabled=True,radius=radius, length=length, color=red)
    ps_cloud.add_vector_quantity("compressive_stress__axes_2_down", -vecs_comp_2, enabled=True,radius=radius, length=length, color=red)
    
    if show : 
        ps.show()
    if clean_after : 
        ps.remove_all_structures()


def display_embryo_graph(Mesh,clean_before = True, clean_after = True,show=True):
    
    ps.init()
    if clean_before : 
        ps.remove_all_structures()
    
    G=Mesh.compute_networkx_graph()
    Edges_for_plotting = []
    for edge in list(G.edges):
        if edge[0]==0 or edge[1] ==1 : 
            continue
        Edges_for_plotting.append(np.array(edge)-1)
    Edges_for_plotting=np.array(Edges_for_plotting)
    Centroids_for_plotting = np.array([a[1] for a in G.nodes.data('centroid')])[1:]
    ps.register_point_cloud("Nodes",Centroids_for_plotting,color=[0,0,0],radius=0.07)
    ps.register_curve_network("Edges",Centroids_for_plotting,Edges_for_plotting,color = [0.5,0.5,0.5],radius=0.02)
    
    
    if show : 
        ps.show()
        
    if clean_after : 
        ps.remove_all_structures()

def display_embryo_graph_forces(G,alpha = 0.05, clean_before = True, clean_after = True,show=True,P0=0, Plot_pressures = True):
    
    ps.init()
    ps.set_ground_plane_mode("none")

    if clean_before : 
        ps.remove_all_structures()
    

    keys_edges = list(G.edges.keys())
    vertices = []
    edges = []
    values = []
    v_p0_list = []
    for i,key in enumerate(keys_edges) : 
        a,b = key

        v2 = G.nodes[b]['centroid']
        if a==0 : 
            v1 = np.mean(G.edges[key]['verts'],axis=0) #+ (np.mean(G.edges[key]['verts'],axis=0)-v2)*1.7
            v_p0_list.append(v1.copy())
        else : 
            v1 = G.nodes[a]['centroid']
        
        vertices.append(v1)
        vertices.append(v2)

        edges.append([2*i,2*i+1])
        values.append(G.edges[key]['tension'])

    vertices = np.array(vertices)
    edges = np.array(edges)
    values = np.array(values)
    v_p0_list = np.array(v_p0_list)

    values = values.clip(np.quantile(values,alpha),np.quantile(values,1-alpha))
    values-=np.amin(values)
    values/=np.amax(values)
    

    colors_tensions = cm.jet(values)[:,:3]

    ps_net = ps.register_curve_network("Graph Edges",vertices,edges,radius=0.007, color = np.array([92,85,141])/255)#,color = "000000")
    ps_net.add_color_quantity("Surface tensions", colors_tensions, defined_on='edges',enabled = True)

    
    Centroids_for_plotting = np.array([a[1] for a in G.nodes.data('centroid')])[np.array(G.nodes.data('volume'))[:,1]>0]#[1:]
    ps_cloud = ps.register_point_cloud("Graph Nodes",Centroids_for_plotting, radius=0.02, color =[0,0,0])# )
    ps_ext = ps.register_point_cloud("Graph Exterior Nodes",v_p0_list,color= cm.magma(1.0)[:3],  radius=0.015)

    if Plot_pressures : 
        Pressure_for_plotting = np.array([a[1] for a in G.nodes.data('pressure')])[np.array(G.nodes.data('volume'))[:,1]>0]#[1:]
        values_p = Pressure_for_plotting.copy()
        values_p-=P0
        
        values_p = values_p.clip(0,np.quantile(values_p,1-alpha))
        #values_p = values_p.clip(np.quantile(values_p,alpha),np.quantile(values_p,1-alpha))
        #print("Extremas of the pressures plotted : ",np.amin(values_p),np.amax(values_p))
        #values_p-=np.amin(values_p)
        values_p/=np.amax(values_p)
        values_p = np.array(1-values_p,dtype=np.float64)

        colors_pressure = cm.magma(values_p)[:,:3]
        ps_cloud.add_color_quantity("Pressure",colors_pressure,enabled = True)
        

        
    if show : 
        ps.show()
        
    if clean_after : 
        ps.remove_all_structures()







####
#Residuals
####

def plot_valid_junctions(Mesh,dict_tensions=None):
    
    if dict_tensions == None: 
        _,dict_tensions,_ = infer_tension(Mesh,mean_tension=1)
    dict_length = Mesh.compute_length_trijunctions()
    dict_validity = {}
    for key in dict_length:
        a,b,c = key
        ga,gb,gc = dict_tensions[(a,b)],dict_tensions[(a,c)],dict_tensions[(b,c)]
        if gc>ga+gb or ga>gc+gb or gb>gc+ga:
            dict_validity[key]=1.0
        else : 
            dict_validity[key]=0.0
    
    ps.set_ground_plane_mode('none')
    ps.remove_all_structures()
    ps.register_surface_mesh("mesh",Mesh.v,Mesh.f[:,:3],color = (1,1,1),transparency = 0.1)
    if np.amax(list(dict_validity.values()))==0:
        plot_trijunctions_topo_change_viewer(Mesh,Dict_trijunctional_values=dict_validity,color="uniform",value_color = (0,1,0),clean_before=False)
    else:
        plot_trijunctions_topo_change_viewer(Mesh,Dict_trijunctional_values=dict_validity,color=1,clean_before=False)


def plot_residual_junctions(Mesh,dict_tensions=None,alpha = 0.05):
    
    if dict_tensions == None: 
        _,dict_tensions,_ = infer_tension(Mesh,mean_tension=1)
    dict_residuals = compute_residual_junctions_dict(Mesh, dict_tensions, alpha = alpha)
    
    print("Extremas of the residuals plotted : ",np.amin(list(dict_residuals.values())),np.amax(list(dict_residuals.values())))
    
    ps.set_ground_plane_mode('none')
    ps.remove_all_structures()
    ps.register_surface_mesh("mesh",Mesh.v,Mesh.f[:,:3],color = (1,1,1),transparency = 0.1)
   
    plot_trijunctions(Mesh,Dict_trijunctional_values=dict_residuals,clean_before=False,cmap = cm.jet)





"""
def plot_force_inference_with_lt(Mesh,force_inference_dicts=None,alpha = 0.05, scalar_quantities = False):
    
    if force_inference_dicts == None : 
        _,dict_tensions,dict_line_tensions,dict_pressure,_=infer_forces_variational_lt(Mesh,mean_tension=1) 
    else : 
        dict_tensions,dict_line_tensions,dict_pressure = force_inference_dicts

    G=Mesh.compute_networkx_graph()
    nx.set_edge_attributes(G, dict_tensions, "tension")
    nx.set_node_attributes(G, dict_pressure, "pressure")

    ps_mesh = view_dict_values_on_mesh(Mesh,dict_tensions,alpha = alpha,ps_mesh = None, clean_before = False, clean_after=False,show=False,name_values = "Surface Tensions")
    ps_mesh = view_pressures_on_mesh(Mesh,dict_pressure,ps_mesh = ps_mesh, alpha = alpha, clean_before=False, clean_after=False, show = False)
    ps_mesh = plot_trijunctions(Mesh,Dict_trijunctional_values=dict_line_tensions,clean_before=False, clean_after=False, show = False,cmap = cm.jet)
    
    plot_stress_tensor(Mesh,G,dict_tensions, dict_pressure, clean_before = False, clean_after = False,show=False)
    display_embryo_graph_forces(G,alpha = alpha, clean_before = False, clean_after = False,show=False,P0 = 0, Plot_pressures = True)
    if scalar_quantities : 
        view_area_derivatives(Mesh, alpha = alpha,ps_mesh = ps_mesh, clean_before = False, clean_after = False,show=False)
        view_volume_derivatives(Mesh, alpha = alpha,ps_mesh = ps_mesh, clean_before = False, clean_after = False,show=False)
        view_mean_curvature_cotan(Mesh, alpha = alpha,ps_mesh = ps_mesh, clean_before = False, clean_after = False,show=False)
        view_mean_curvature_robust(Mesh, alpha = alpha,ps_mesh = ps_mesh, clean_before = False, clean_after = False,show=False)
        view_gaussian_curvature(Mesh, alpha = alpha,ps_mesh = ps_mesh, clean_before = False, clean_after = False,show=False)
        view_discrepancy_of_principal_curvatures(Mesh, alpha = alpha,ps_mesh = ps_mesh, clean_before = False, clean_after = False,show=False)

    ps.show()

from dw3d.Curvature import compute_curvature_vertices_robust_laplacian

def view_mean_curvature_robust(Mesh,alpha = 0.05,ps_mesh = None, remove_trijunctions=True,clean_before = True, clean_after = True,show=True,scattered = False):
    
    H,_,_=compute_curvature_vertices_robust_laplacian(Mesh)
    
    vmin, vmax = np.quantile(H,alpha),np.quantile(H,1-alpha)
    print("Extremas of mean curvature (robust) plotted: ",vmin,vmax)
    H = H.clip(vmin,vmax)

    ps_mesh = view_vertex_values_on_embryo(Mesh,H,ps_mesh = ps_mesh,name_values="Mean Curvature Robust",remove_trijunctions=remove_trijunctions,clean_before = clean_before, clean_after = clean_after,show=show, scattered = scattered)
        
    return(ps_mesh)

"""


import numpy as np 
from scipy import linalg
from .Pressure_inference import infer_pressure


def infer_forces(Mesh, mean_tension=1, P0 = 0, mode_tension = "Young-Dupré", mode_pressure = "Variational"): 
    _,dict_tensions,_ = infer_tension(Mesh, mean_tension=mean_tension, mode = mode_tension)
    _,dict_pressures,_ = infer_pressure(Mesh,dict_tensions,mode=mode_pressure, P0 = P0)
    return(dict_tensions, dict_pressures)

def infer_forces_variational_lt(Mesh):
    #TODO
    return(None)

def change_idx_cells(faces,mapping):
    #mapping : {key_init:key_end} has to be a bijection
    new_faces = faces.copy()
    for key in mapping : 
        new_faces[faces[:,3]==key][x:,3]=mapping[key]
        new_faces[faces[:,4]==key][:,4]=mapping[key]
    return(new_faces)



def cot(x): 
    return(np.cos(x)/np.sin(x))
def compute_z(phi1,phi2,phi3): 
    z1 = np.sin(phi1)/(1-np.cos(phi1))
    z2 = np.sin(phi2)/(1-np.cos(phi2))
    z3 = np.sin(phi3)/(1-np.cos(phi3))
    return(z1,z2,z3)

def factors_z(z1,z2,z3): 
    f1 = z2*z3
    f2 = z1*z3
    f3 = z1*z2
    s1 = z1*(z2 + z3)/2
    s2 = z2*(z1 + z3)/2
    s3 = z3*(z1 + z2)/2
    return(f1,f2,f3,s1,s2,s3)






"""
FORCE INFERENCE ONLY TAKING INTO ACCOUNT TRIJUNCTIONS : 
THERE ARE A LOT OF QUADRIJUNCTIONS IN THE MESH 
HOWEVER, TAKING ONLY THE TRIJUNCTIONS INTO ACCOUNT SHOULD SUFFICE IN MOST CASES, ESPECIALLY WITH A FEW CELLS
BECAUSE : 
THE PROBLEM IS HIGHLY OVERDETERMINED
AND THE QUADRIJUNCTIONS ARE NOT VERY PRESENT (THEY OCCUPY A LITTLE LENGTH)
"""



def infer_tension(Mesh,mean_tension=1,mode='Young-Dupré'):

    if mode == 'Young-Dupré': 
        return(infer_tension_symmetrical_yd(Mesh,mean_tension))
    elif mode == 'Eq': 
        return(infer_tension_equilibrium(Mesh,mean_tension))
    elif mode == 'Projection Young-Dupré': 
        return(infer_tension_projection_yd(Mesh,mean_tension))
    elif mode == 'cotan': 
        return(infer_tension_cotan(Mesh,mean_tension))
    elif mode == 'inv_cotan':
        return(infer_tension_inv_cotan(Mesh,mean_tension))
    elif mode =='Lami': 
        return(infer_tension_lamy(Mesh,mean_tension))
    elif mode =='Lami inverse': 
        return(infer_tension_inv_lamy(Mesh,mean_tension))
    elif mode =='Lami logarithm': 
        return(infer_tension_lamy_log(Mesh,mean_tension))
    elif mode =='Variational':
        return(infer_tension_variational_yd(Mesh,mean_tension))
        
    else : 
        print("Unimplemented method")
        return(None)




def build_matrix_tension_inv_lamy(dict_rad,dict_areas,dict_length_tri,n_materials,mean_tension=1):

    Total_length_mesh = sum(dict_length_tri.values())/3 #total length of all the junctions on the mesh (on which are evaluated the surface tensions)

    #Index interfaces :
    T = np.array(sorted(list(dict_areas.keys())))
    kt = np.amax(T)+1
    Keys_t = T[:,0]*kt + T[:,1]
    reverse_map_t = dict(zip(Keys_t, np.arange(len(Keys_t))))
    
    #Index angles : 
    A = np.array(list(dict_rad.keys()))
    A = -np.sort(-A,axis=1)
    ka = np.amax(A)
    Keys_a = A[:,0]*ka**2 + A[:,1]*ka + A[:,2]
    Keys_a,index = np.unique(Keys_a,return_index=True)
    A=A[index]
    
    #MX = B
    #x : structure : [0,nm[ : tensions
    nm = len(T)
    nc = n_materials-1
    nj = len(A)
    size = (3*nj + 1)*(nm)
    B = np.zeros(3*nj + 1)
    B[-1]=nm*mean_tension  #Sum of tensions = number of membrane <=> mean of tensions = 1
    M = np.zeros((3*nj + 1, nm))
    
    
    #CLASSICAL 
    #YOUNG-DUPRÉ
    for i,key in enumerate(Keys_a):
        c = key//ka**2
        b = (key-(c*(ka**2)))//ka
        a = key%ka

        O_b = dict_rad[(a,b,c)] #O_b = abc angle
        O_a = dict_rad[(b,a,c)] #O_a = cab angle
        O_c = dict_rad[(a,c,b)]
        idx_tab = reverse_map_t[min(a,b)*kt+max(a,b)]
        idx_tbc = reverse_map_t[min(b,c)*kt+max(b,c)]
        idx_tac = reverse_map_t[min(a,c)*kt+max(a,c)]

        factor = dict_length_tri[(a,b,c)]/Total_length_mesh
        M[3*i,idx_tab]=np.sin(O_a)    * factor
        M[3*i,idx_tbc]=-np.sin(O_c)   * factor

        M[3*i+1,idx_tab]=np.sin(O_b)  * factor
        M[3*i+1,idx_tac]=-np.sin(O_c) * factor

        M[3*i+2,idx_tbc]=np.sin(O_b)  * factor
        M[3*i+2,idx_tac]=-np.sin(O_a) * factor

    M[3*nj,:nm]=1 #Enforces mean of tensions = 1 (see B[-1])

    return(M,B)





def build_matrix_tension_lamy(dict_rad,dict_areas,dict_length_tri,n_materials,mean_tension=1):

    Total_length_mesh = sum(dict_length_tri.values())/3 #total length of all the junctions on the mesh (on which are evaluated the surface tensions)

    #Index interfaces :
    T = np.array(sorted(list(dict_areas.keys())))
    kt = np.amax(T)+1
    Keys_t = T[:,0]*kt + T[:,1]
    reverse_map_t = dict(zip(Keys_t, np.arange(len(Keys_t))))
    
    #Index angles : 
    A = np.array(list(dict_rad.keys()))
    A = -np.sort(-A,axis=1)
    ka = np.amax(A)
    Keys_a = A[:,0]*ka**2 + A[:,1]*ka + A[:,2]
    Keys_a,index = np.unique(Keys_a,return_index=True)
    A=A[index]
    
    #MX = B
    #x : structure : [0,nm[ : tensions
    nm = len(T)
    nc = n_materials-1
    nj = len(A)
    size = (3*nj + 1)*(nm)
    B = np.zeros(3*nj + 1)
    B[-1]=nm*mean_tension  #Sum of tensions = number of membrane <=> mean of tensions = 1
    M = np.zeros((3*nj + 1, nm))
    
    
    #CLASSICAL 
    #YOUNG-DUPRÉ
    for i,key in enumerate(Keys_a):
        c = key//ka**2
        b = (key-(c*(ka**2)))//ka
        a = key%ka

        O_b = dict_rad[(a,b,c)] #O_b = abc angle
        O_a = dict_rad[(b,a,c)] #O_a = cab angle
        O_c = dict_rad[(a,c,b)]
        idx_tab = reverse_map_t[min(a,b)*kt+max(a,b)]
        idx_tbc = reverse_map_t[min(b,c)*kt+max(b,c)]
        idx_tac = reverse_map_t[min(a,c)*kt+max(a,c)]

        factor = dict_length_tri[(a,b,c)]/Total_length_mesh
        M[3*i,idx_tab]=1/np.sin(O_c)    * factor
        M[3*i,idx_tbc]=-1/np.sin(O_a)   * factor

        M[3*i+1,idx_tab]=1/np.sin(O_c)  * factor
        M[3*i+1,idx_tac]=-1/np.sin(O_b) * factor

        M[3*i+2,idx_tbc]=1/np.sin(O_a)  * factor
        M[3*i+2,idx_tac]=-1/np.sin(O_b) * factor

    M[3*nj,:nm]=1 #Enforces mean of tensions = 1 (see B[-1])

    return(M,B)


def build_matrix_tension_projection_yd(dict_rad,dict_areas,dict_length_tri,n_materials,mean_tension=1):

    Total_length_mesh = sum(dict_length_tri.values())/3 #total length of all the junctions on the mesh (on which are evaluated the surface tensions)

    #Index interfaces :
    T = np.array(sorted(list(dict_areas.keys())))
    kt = np.amax(T)+1
    Keys_t = T[:,0]*kt + T[:,1]
    reverse_map_t = dict(zip(Keys_t, np.arange(len(Keys_t))))
    
    #Index angles : 
    A = np.array(list(dict_rad.keys()))
    A = -np.sort(-A,axis=1)
    ka = np.amax(A)
    Keys_a = A[:,0]*ka**2 + A[:,1]*ka + A[:,2]
    Keys_a,index = np.unique(Keys_a,return_index=True)
    A=A[index]
    
    #MX = B
    #x : structure : [0,nm[ : tensions
    nm = len(T)
    nc = n_materials-1
    nj = len(A)
    size = (2*nj + 1)*(nm)
    B = np.zeros(2*nj + 1)
    B[-1]=nm*mean_tension  #Sum of tensions = number of membrane <=> mean of tensions = 1
    M = np.zeros((2*nj + 1, nm))
    
    
    #CLASSICAL 
    #YOUNG-DUPRÉ
    for i,key in enumerate(Keys_a):
        c = key//ka**2
        b = (key-(c*(ka**2)))//ka
        a = key%ka
        O_b = dict_rad[(a,b,c)] #O_b = abc angle
        O_a = dict_rad[(b,a,c)] #O_a = cab angle
        idx_tab = reverse_map_t[min(a,b)*kt+max(a,b)]
        idx_tbc = reverse_map_t[min(b,c)*kt+max(b,c)]
        idx_tac = reverse_map_t[min(a,c)*kt+max(a,c)]

        factor = dict_length_tri[(a,b,c)]/Total_length_mesh
        M[2*i,idx_tab]=1              * factor
        M[2*i,idx_tbc]=np.cos(O_b)    * factor
        M[2*i,idx_tac]=np.cos(O_a)    * factor
        M[2*i+1,idx_tac]=-np.sin(O_a) * factor
        M[2*i+1,idx_tbc]=np.sin(O_b)  * factor

    M[2*nj,:nm]=1 #Enforces mean of tensions = 1 (see B[-1])

    return(M,B)









def build_matrix_tension_lamy_log(dict_rad,dict_areas,dict_length_tri,n_materials,mean_tension=1):

    Total_length_mesh = sum(dict_length_tri.values())/3 #total length of all the junctions on the mesh (on which are evaluated the surface tensions)

    #Index interfaces :
    T = np.array(sorted(list(dict_areas.keys())))
    kt = np.amax(T)+1
    Keys_t = T[:,0]*kt + T[:,1]
    reverse_map_t = dict(zip(Keys_t, np.arange(len(Keys_t))))
    
    #Index angles : 
    A = np.array(list(dict_rad.keys()))
    A = -np.sort(-A,axis=1)
    ka = np.amax(A)
    Keys_a = A[:,0]*ka**2 + A[:,1]*ka + A[:,2]
    Keys_a,index = np.unique(Keys_a,return_index=True)
    A=A[index]
    
    #MX = B
    #x : structure : [0,nm[ : tensions
    nm = len(T)
    nc = n_materials-1
    nj = len(A)
    size = (3*nj + 1)*(nm)
    B = np.zeros(3*nj + 1)
    B[-1]=nm*np.log(mean_tension)  #Sum of tensions = number of membrane <=> mean of tensions = 1
    M = np.zeros((3*nj + 1, nm))
    
    
    #CLASSICAL 
    #YOUNG-DUPRÉ
    for i,key in enumerate(Keys_a):
        c = key//ka**2
        b = (key-(c*(ka**2)))//ka
        a = key%ka

        O_b = dict_rad[(a,b,c)] #O_b = abc angle
        O_a = dict_rad[(b,a,c)] #O_a = cab angle
        O_c = dict_rad[(a,c,b)]
        idx_tab = reverse_map_t[min(a,b)*kt+max(a,b)]
        idx_tbc = reverse_map_t[min(b,c)*kt+max(b,c)]
        idx_tac = reverse_map_t[min(a,c)*kt+max(a,c)]

        factor = dict_length_tri[(a,b,c)]/Total_length_mesh
        M[3*i,idx_tab]=1    * factor
        M[3*i,idx_tbc]=-1   * factor
        B[3*i] = (np.log(np.sin(O_c)) - np.log(np.sin(O_a)))* factor

        M[3*i+1,idx_tab]=1    * factor
        M[3*i+1,idx_tac]=-1   * factor
        B[3*i+1] = (np.log(np.sin(O_c)) - np.log(np.sin(O_b)))* factor

        M[3*i+2,idx_tbc]=1    * factor
        M[3*i+2,idx_tac]=-1   * factor
        B[3*i+2] = (np.log(np.sin(O_a)) - np.log(np.sin(O_b)))* factor

    M[3*nj,:nm]=1 #Enforces mean of tensions = 1 (see B[-1])

    return(M,B)



def build_matrix_tension_cotan(dict_rad,dict_areas,dict_length_tri,n_materials,mean_tension=1):

    Total_length_mesh = sum(dict_length_tri.values())/3 #total length of all the junctions on the mesh (on which are evaluated the surface tensions)

    #Index interfaces :
    T = np.array(sorted(list(dict_areas.keys())))
    kt = np.amax(T)+1
    Keys_t = T[:,0]*kt + T[:,1]
    reverse_map_t = dict(zip(Keys_t, np.arange(len(Keys_t))))
    
    #Index angles : 
    A = np.array(list(dict_rad.keys()))
    A = -np.sort(-A,axis=1)
    ka = np.amax(A)
    Keys_a = A[:,0]*ka**2 + A[:,1]*ka + A[:,2]
    Keys_a,index = np.unique(Keys_a,return_index=True)
    A=A[index]
    
    #MX = B
    #x : structure : [0,nm[ : tensions
    nm = len(T)
    nc = n_materials-1
    nj = len(A)
    size = (3*nj + 1)*(nm)
    B = np.zeros(3*nj + 1)
    B[-1]=nm*mean_tension  #Sum of tensions = number of membrane <=> mean of tensions = 1
    M = np.zeros((3*nj + 1, nm))
    
    
    #CLASSICAL 
    #YOUNG-DUPRÉ
    for i,key in enumerate(Keys_a):
        c = key//ka**2
        b = (key-(c*(ka**2)))//ka
        a = key%ka

        O_b = dict_rad[(a,b,c)] #O_b = abc angle
        O_a = dict_rad[(b,a,c)] #O_a = cab angle
        O_c = dict_rad[(a,c,b)]
        idx_tab = reverse_map_t[min(a,b)*kt+max(a,b)]
        idx_tbc = reverse_map_t[min(b,c)*kt+max(b,c)]
        idx_tac = reverse_map_t[min(a,c)*kt+max(a,c)]

        z1,z2,z3 = compute_z(O_a,O_b,O_c)
        f1,f2,f3,s1,s2,s3 = factors_z(z1,z2,z3)

        factor = dict_length_tri[(a,b,c)]/Total_length_mesh
        M[3*i,idx_tab]=-s3-f3    * factor
        M[3*i,idx_tbc]=s3        * factor
        M[3*i,idx_tac]=s3        * factor

        M[3*i+1,idx_tab]=s1      * factor
        M[3*i+1,idx_tbc]=-s1-f1  * factor
        M[3*i+1,idx_tac]=s1      * factor

        M[3*i+2,idx_tab]=s2      * factor
        M[3*i+2,idx_tbc]=s2      * factor
        M[3*i+2,idx_tac]=-s2-f2  * factor

    M[3*nj,:nm]=1 #Enforces mean of tensions = 1 (see B[-1])

    return(M,B)


def build_matrix_tension_inv_cotan(dict_rad,dict_areas,dict_length_tri,n_materials,mean_tension=1):

    Total_length_mesh = sum(dict_length_tri.values())/3 #total length of all the junctions on the mesh (on which are evaluated the surface tensions)

    #Index interfaces :
    T = np.array(sorted(list(dict_areas.keys())))
    kt = np.amax(T)+1
    Keys_t = T[:,0]*kt + T[:,1]
    reverse_map_t = dict(zip(Keys_t, np.arange(len(Keys_t))))
    
    #Index angles : 
    A = np.array(list(dict_rad.keys()))
    A = -np.sort(-A,axis=1)
    ka = np.amax(A)
    Keys_a = A[:,0]*ka**2 + A[:,1]*ka + A[:,2]
    Keys_a,index = np.unique(Keys_a,return_index=True)
    A=A[index]
    
    #MX = B
    #x : structure : [0,nm[ : tensions
    nm = len(T)
    nc = n_materials-1
    nj = len(A)
    size = (3*nj + 1)*(nm)
    B = np.zeros(3*nj + 1)
    B[-1]=nm*mean_tension  #Sum of tensions = number of membrane <=> mean of tensions = 1
    M = np.zeros((3*nj + 1, nm))
    
    
    #CLASSICAL 
    #YOUNG-DUPRÉ
    for i,key in enumerate(Keys_a):
        c = key//ka**2
        b = (key-(c*(ka**2)))//ka
        a = key%ka

        O_b = dict_rad[(a,b,c)] #O_b = abc angle
        O_a = dict_rad[(b,a,c)] #O_a = cab angle
        O_c = dict_rad[(a,c,b)]
        idx_tab = reverse_map_t[min(a,b)*kt+max(a,b)]
        idx_tbc = reverse_map_t[min(b,c)*kt+max(b,c)]
        idx_tac = reverse_map_t[min(a,c)*kt+max(a,c)]

        z1,z2,z3 = compute_z(O_a,O_b,O_c)
        f1,f2,f3,s1,s2,s3 = factors_z(z1,z2,z3)

        factor = dict_length_tri[(a,b,c)]/Total_length_mesh
        M[3*i,idx_tab]=-1/s3-1/f3    * factor
        M[3*i,idx_tbc]=1/f3        * factor
        M[3*i,idx_tac]=1/f3        * factor

        M[3*i+1,idx_tab]=1/f1      * factor
        M[3*i+1,idx_tbc]=-1/s1-1/f1  * factor
        M[3*i+1,idx_tac]=1/f1      * factor

        M[3*i+2,idx_tab]=1/f2      * factor
        M[3*i+2,idx_tbc]=1/f2      * factor
        M[3*i+2,idx_tac]=-1/f2-1/s2  * factor

    M[3*nj,:nm]=1 #Enforces mean of tensions = 1 (see B[-1])

    return(M,B)






def build_matrix_equilibrium_tensions(Mesh,dict_areas,mean_tension=1):

    

    #Index interfaces :
    T = np.array(sorted(list(dict_areas.keys())))
    kt = np.amax(T)+1
    Keys_t = T[:,0]*kt + T[:,1]
    reverse_map_t = dict(zip(Keys_t, np.arange(len(Keys_t))))
    
    #MX = B
    #x : structure : [0,nm[ : tensions
    nm = len(T)
    table=[]
    for edge in Mesh.half_edges : 
        if len(edge.twin)>1 : 
            table.append(edge.key)
    ne = len(table)
    
    B = np.zeros(3*ne + 1)
    B[-1]=nm*mean_tension  #Sum of tensions = number of membrane <=> mean of tensions = 1
    M = np.zeros((3*ne+1, nm))
    M[-1]=1

    ne_i = 0
    for edge in Mesh.half_edges : 
        if len(edge.twin)==1 : 
            continue
            
            
        list_edges = [edge.key]+edge.twin
        vectors_tension={}
        for current_edge_idx in list_edges:
            #print(current_edge_idx)
            current_edge = Mesh.half_edges[current_edge_idx]
            #face_attached = current_edge.incident_face
            #Find the edges of the faces
            e = current_edge.incident_face.outer_component#current_edge
            edges_face = []
            for i in range(3): 
                #print(e.incident_face,e)
                edges_face.append([e.origin.key,e.destination.key])
                e = e.next
            edges_face=np.array(edges_face)

            #find the direction of the force vector
            verts_face = Mesh.v[np.unique(edges_face)]
            verts_edge = Mesh.v[[current_edge.origin.key,current_edge.destination.key]]

            index_edge = []
            index_not = []
            for index in (np.unique(edges_face)): 
                if index in [current_edge.origin.key,current_edge.destination.key] : 
                    index_edge.append(index)
                else : 
                    index_not.append(index)


            #np.unique(edges_face),index,[edge.origin.key,edge.destination.key]
            #index_edge,index_not
            vector_tension = Mesh.v[index_not[0]]-Mesh.v[index_edge[0]]
            edge_vect = verts_edge[1]-verts_edge[0]
            edge_vect/=np.linalg.norm(edge_vect)
            vector_tension-=np.dot(vector_tension,edge_vect)*edge_vect
            vector_tension/=np.linalg.norm(vector_tension)
            
            a,b = current_edge.incident_face.material_1,current_edge.incident_face.material_2
            idx_tension = reverse_map_t[min(a,b)*kt+max(a,b)]
            
            M[3*ne_i,idx_tension]=vector_tension[0]
            M[3*ne_i+1,idx_tension]=vector_tension[1]
            M[3*ne_i+2,idx_tension]=vector_tension[2]
            
        ne_i+=1
        
    return(M,B)

def cot(x): 
    return(np.cos(x)/np.sin(x))
def tensions_equilibrium(phi1,phi2,phi3):
    t1 = cot(phi1/2)*(cot(phi2/2) + cot(phi3/2))
    t2 = cot(phi2/2)*(cot(phi1/2) + cot(phi3/2))
    t3 = cot(phi3/2)*(cot(phi1/2) + cot(phi2/2))
    return(t1,t2,t3)


def build_matrix_tension_symmetrical_yd(dict_rad,dict_areas,dict_length_tri,n_materials,mean_tension=1):

    Total_length_mesh = sum(dict_length_tri.values())/3 #total length of all the junctions on the mesh (on which are evaluated the surface tensions)

    #Index interfaces :
    T = np.array(sorted(list(dict_areas.keys())))
    kt = np.amax(T)+1
    Keys_t = T[:,0]*kt + T[:,1]
    reverse_map_t = dict(zip(Keys_t, np.arange(len(Keys_t))))
    
    #Index angles : 
    A = np.array(list(dict_rad.keys()))
    A = -np.sort(-A,axis=1)
    ka = np.amax(A)
    Keys_a = A[:,0]*ka**2 + A[:,1]*ka + A[:,2]
    Keys_a,index = np.unique(Keys_a,return_index=True)
    A=A[index]
    
    #MX = B
    #x : structure : [0,nm[ : tensions
    nm = len(T)
    nc = n_materials-1
    nj = len(A)
    size = (3*nj + 1)*(nm)
    B = np.zeros(3*nj + 1)
    B[-1]=nm*mean_tension  #Sum of tensions = number of membrane <=> mean of tensions = 1
    M = np.zeros((3*nj + 1, nm))
    
    
    #CLASSICAL 
    #YOUNG-DUPRÉ
    for i,key in enumerate(Keys_a):
        c = key//ka**2
        b = (key-(c*(ka**2)))//ka
        a = key%ka

        O_b = dict_rad[(a,b,c)] #O_b = abc angle
        O_a = dict_rad[(b,a,c)] #O_a = cab angle
        O_c = dict_rad[(a,c,b)] #O_c = acb angle
        
        idx_tab = reverse_map_t[min(a,b)*kt+max(a,b)]
        idx_tbc = reverse_map_t[min(b,c)*kt+max(b,c)]
        idx_tac = reverse_map_t[min(a,c)*kt+max(a,c)]

        factor = dict_length_tri[(a,b,c)]/Total_length_mesh
        M[3*i,idx_tab]=1             * factor
        M[3*i,idx_tbc]=np.cos(O_b)   * factor
        M[3*i,idx_tac]=np.cos(O_a)   * factor

        M[3*i+1,idx_tab]=np.cos(O_b)  * factor
        M[3*i+1,idx_tbc]=1            * factor
        M[3*i+1,idx_tac]=np.cos(O_c)  * factor

        M[3*i+2,idx_tab]=np.cos(O_a)             * factor
        M[3*i+2,idx_tbc]=np.cos(O_c)  * factor
        M[3*i+2,idx_tac]=1               * factor

    M[3*nj,:nm]=1 #Enforces mean of tensions = 1 (see B[-1])

    return(M,B)


def build_matrix_discrete(DA,DV,materials,mean_tension=1):
    
    T = np.array(list(DA.keys()))
    kt = np.amax(T)+1
    Keys_t = T[:,0] + T[:,1]*kt
    reverse_map_t = dict(zip(Keys_t, np.arange(len(Keys_t))))

    #MX = B
    #x : structure : ([0,nm[ : tensions) ([nm:nm+nc] : pressions) : (ya,yb,yc,yd...p0,p1,p2..pnc)
    nm = len(T)
    nc = len(materials)-1
    size = (nm + nc +1)*(nm +nc +1)
    B = np.zeros(nm + nc +1)
    B[-1]=nm*mean_tension #Sum of tensions = number of membrane <=> mean of tensions = mean_tensions
    M = np.zeros((nm + nc +1, nm +nc))
    #print("n_j",nj,"\nn_m",nm,"\nn_c",nc,"\nNumber of unknowns ",nm+nc+1)

    ###VARIATIONAL
    ###YOUNG-DUPRÉ
    ###EQUATIONS
    
    for i,tup_l in enumerate(DA.keys()): 
        a_l,b_l = tup_l
        key_l = a_l + b_l*kt
        
        for tup_m in DA.keys(): 
            a_m,b_m = tup_m
            key_m = a_m + b_m*kt
            idx_tm = reverse_map_t[key_m]
            M[i,idx_tm]=np.sum(DA[tup_m]*DA[tup_l])
            
        for n,key_n in enumerate(materials[1:]): 
            M[i,nm+n] = -np.sum(DV[key_n]*DA[tup_l])
        
    off=nm 
    
    ###VARIATIONAL
    ###LAPLACE
    ###EQUATIONS
    for i,key_i in enumerate(materials[1:]): 

        for tup in DA.keys(): 
            a,b = tup
            key = a + b*kt
            idx_tk = reverse_map_t[key]
            M[off+i,idx_tk] = np.sum(DA[(a,b)]*DV[key_i])
            
        for n,key_n in enumerate(materials[1:]): 
            M[off+i,nm+n] = -np.sum(DV[key_n]*DV[key_i])

    off+=nc
    
    Lines_sum = np.sum(M,axis=1)
    for i in range(len(M)-1): 
        if Lines_sum[i]!=0 : 
            M[i]/=Lines_sum[i]
    
    M[off,:nm]=1 #Enforces mean of tensions = mean_tensions (see B[-1])
    
    return(M,B,nm)

def extract_submatrices(M,nm): 
    F_g = M[:nm,:nm]
    F_p = - M[:nm,nm:]
    G_g = M[nm:-1,:nm]
    G_p = -M[nm:-1,nm:]
    return(F_g,F_p,G_g,G_p)













def infer_tension_projection_yd(Mesh,mean_tension=1):
    dict_rad,_,dict_length = Mesh.compute_angles_tri(unique=False)
    dict_areas = Mesh.compute_areas_interfaces()
    M,B = build_matrix_tension_projection_yd(dict_rad,dict_areas,dict_length,Mesh.n_materials,mean_tension)
    x, resid, rank, sigma = linalg.lstsq(M, B)
    dict_tensions = {}
    T = np.array(sorted(list(dict_areas.keys())))
    nm = len(T)    
    for i in range(nm): 
        dict_tensions[tuple(T[i])]=x[i]
    
    
    return(x,dict_tensions,resid)


def infer_tension_symmetrical_yd(Mesh,mean_tension=1):
    dict_rad,_,dict_length = Mesh.compute_angles_tri(unique=False)
    dict_areas = Mesh.compute_areas_interfaces()
    M,B = build_matrix_tension_symmetrical_yd(dict_rad,dict_areas,dict_length,Mesh.n_materials,mean_tension)
    x, resid, rank, sigma = linalg.lstsq(M, B)
    dict_tensions = {}
    T = np.array(sorted(list(dict_areas.keys())))
    nm = len(T)    
    for i in range(nm): 
        dict_tensions[tuple(T[i])]=x[i]
    
    
    return(x,dict_tensions,resid)

def infer_tension_cotan(Mesh,mean_tension=1):
    dict_rad,_,dict_length = Mesh.compute_angles_tri(unique=False)
    dict_areas = Mesh.compute_areas_interfaces()
    M,B = build_matrix_tension_cotan(dict_rad,dict_areas,dict_length,Mesh.n_materials,mean_tension)
    x, resid, rank, sigma = linalg.lstsq(M, B)
    dict_tensions = {}
    T = np.array(sorted(list(dict_areas.keys())))
    nm = len(T)    
    for i in range(nm): 
        dict_tensions[tuple(T[i])]=x[i]
    
    
    return(x,dict_tensions,resid)


def infer_tension_inv_cotan(Mesh,mean_tension=1):
    dict_rad,_,dict_length = Mesh.compute_angles_tri(unique=False)
    dict_areas = Mesh.compute_areas_interfaces()
    M,B = build_matrix_tension_inv_cotan(dict_rad,dict_areas,dict_length,Mesh.n_materials,mean_tension)
    x, resid, rank, sigma = linalg.lstsq(M, B)
    dict_tensions = {}
    T = np.array(sorted(list(dict_areas.keys())))
    nm = len(T)    
    for i in range(nm): 
        dict_tensions[tuple(T[i])]=x[i]
    
    
    return(x,dict_tensions,resid)


def infer_tension_lamy(Mesh,mean_tension=1):
    dict_rad,_,dict_length = Mesh.compute_angles_tri(unique=False)
    dict_areas = Mesh.compute_areas_interfaces()
    M,B = build_matrix_tension_lamy(dict_rad,dict_areas,dict_length,Mesh.n_materials,mean_tension)
    x, resid, rank, sigma = linalg.lstsq(M, B)
    dict_tensions = {}
    T = np.array(sorted(list(dict_areas.keys())))
    nm = len(T)    
    for i in range(nm): 
        dict_tensions[tuple(T[i])]=x[i]
    
    
    return(x,dict_tensions,resid)

def infer_tension_inv_lamy(Mesh,mean_tension=1):
    dict_rad,_,dict_length = Mesh.compute_angles_tri(unique=False)
    dict_areas = Mesh.compute_areas_interfaces()
    M,B = build_matrix_tension_inv_lamy(dict_rad,dict_areas,dict_length,Mesh.n_materials,mean_tension)
    x, resid, rank, sigma = linalg.lstsq(M, B)
    dict_tensions = {}
    T = np.array(sorted(list(dict_areas.keys())))
    nm = len(T)    
    for i in range(nm): 
        dict_tensions[tuple(T[i])]=x[i]
    
    
    return(x,dict_tensions,resid)



def infer_tension_lamy_log(Mesh,mean_tension=1):
    dict_rad,_,dict_length = Mesh.compute_angles_tri(unique=False)
    dict_areas = Mesh.compute_areas_interfaces()
    M,B = build_matrix_tension_lamy_log(dict_rad,dict_areas,dict_length,Mesh.n_materials,mean_tension)
    x_log, resid, rank, sigma = linalg.lstsq(M, B)
    x = np.exp(x_log)
    dict_tensions = {}
    T = np.array(sorted(list(dict_areas.keys())))
    nm = len(T)    
    for i in range(nm): 
        dict_tensions[tuple(T[i])]=x[i]
    
    return(x,dict_tensions,resid)








def infer_tension_variational_yd(Mesh,mean_tension=1,prints = False):

    DA,DV = Mesh.compute_area_derivatives(),Mesh.compute_volume_derivatives()
    
    M,B,nm= build_matrix_discrete(DA,DV,Mesh.materials)
    
    F_g,F_p,G_g,G_p=extract_submatrices(M,nm)
    
    

    Mat_t = F_g - F_p@(np.linalg.inv(G_p))@G_g
    if prints : 
        print(F_g.shape,F_p.shape,G_g.shape,G_p.shape)
        print("Det of F_g:",np.linalg.det(F_g))
        print("Det of G_p:",np.linalg.det(G_p))

    Mat_inference = np.zeros((nm+1,nm))
    Mat_inference[:nm,:nm]=Mat_t
    Mat_inference[nm:]=1
    B_inf = np.zeros(nm+1)
    B_inf[-1]=nm
    x, resid, rank, sigma = linalg.lstsq(Mat_inference, B_inf)
    
    dict_tensions = {}
    dict_pressures = {0:0}
    
    T = np.array(list(DA.keys()))
    nm = len(T)
    
    for i in range(nm): 
        dict_tensions[tuple(T[i])]=x[i]
        
    P=np.linalg.inv(G_p)@G_g@x
        
    
    return(x,dict_tensions,resid)











def infer_tension_equilibrium(Mesh,mean_tension=1):
    dict_areas = Mesh.compute_areas_interfaces()
    M,B = build_matrix_equilibrium_tensions(Mesh,dict_areas,mean_tension)
    x, resid, rank, sigma = linalg.lstsq(M, B)
    dict_tensions = {}
    T = np.array(sorted(list(dict_areas.keys())))
    nm = len(T)    
    for i in range(nm): 
        dict_tensions[tuple(T[i])]=x[i]

    return(x,dict_tensions,resid)






###RESIDUAL STUFF

def build_matrix_tension_validity(dict_rad,dict_areas,dict_length_tri,n_materials,mean_tension=1):

    Total_length_mesh = sum(dict_length_tri.values())/3 #total length of all the junctions on the mesh (on which are evaluated the surface tensions)

    #Index interfaces :
    T = np.array(sorted(list(dict_areas.keys())))
    kt = np.amax(T)+1
    Keys_t = T[:,0]*kt + T[:,1]
    reverse_map_t = dict(zip(Keys_t, np.arange(len(Keys_t))))
    
    #Index angles : 
    A = np.array(list(dict_rad.keys()))
    A = -np.sort(-A,axis=1)
    ka = np.amax(A)
    Keys_a = A[:,0]*ka**2 + A[:,1]*ka + A[:,2]
    Keys_a,index = np.unique(Keys_a,return_index=True)
    A=A[index]
    
    #MX = B
    #x : structure : [0,nm[ : tensions
    nm = len(T)
    nc = n_materials-1
    nj = len(A)
    size = (3*nj + 1)*(nm)
    B = np.zeros(3*nj + 1)
    B[-1]=nm*mean_tension  #Sum of tensions = number of membrane <=> mean of tensions = 1
    M = np.zeros((3*nj + 1, nm))
    
    
    #CLASSICAL 
    #YOUNG-DUPRÉ
    for i,key in enumerate(Keys_a):
        c = key//ka**2
        b = (key-(c*(ka**2)))//ka
        a = key%ka

        O_b = dict_rad[(a,b,c)] #O_b = abc angle
        O_a = dict_rad[(b,a,c)] #O_a = cab angle
        O_c = dict_rad[(a,c,b)] #O_c = acb angle
        
        idx_tab = reverse_map_t[min(a,b)*kt+max(a,b)]
        idx_tbc = reverse_map_t[min(b,c)*kt+max(b,c)]
        idx_tac = reverse_map_t[min(a,c)*kt+max(a,c)]

        M[3*i,idx_tab]=1             
        M[3*i,idx_tbc]=np.cos(O_b)   
        M[3*i,idx_tac]=np.cos(O_a)   

        M[3*i+1,idx_tab]=np.cos(O_b)  
        M[3*i+1,idx_tbc]=1            
        M[3*i+1,idx_tac]=np.cos(O_c)  

        M[3*i+2,idx_tab]=np.cos(O_a)             
        M[3*i+2,idx_tbc]=np.cos(O_c)  
        M[3*i+2,idx_tac]=1               

    M[3*nj,:nm]=1 #Enforces mean of tensions = 1 (see B[-1])

    return(M,B)

def compute_residual_junctions_dict(Mesh,dict_tensions, alpha = 0.05):
    dict_rad,_,dict_length = Mesh.compute_angles_tri(unique=False)
    dict_areas = Mesh.compute_areas_interfaces()
    M,B = build_matrix_tension_validity(dict_rad,dict_areas,dict_length,Mesh.n_materials,mean_tension=1)
    
    T = np.array(sorted(list(dict_areas.keys())))
    nm = len(T)  
    x=np.zeros(nm)
    for i in range(nm): 
        x[i]=dict_tensions[tuple(T[i])]
    
    A = np.array(sorted(list(dict_rad.keys())))
    A = -np.sort(-A,axis=1)
    ka = np.amax(A)
    Keys_a = A[:,0]*ka**2 + A[:,1]*ka + A[:,2]
    Keys_a,index = np.unique(Keys_a,return_index=True)
    A=A[index]
    dict_residuals = {}
    array_resid = (np.abs(M@x-B)**2)[:-1]
    
    array_resid = array_resid.clip(np.quantile(array_resid,alpha),np.quantile(array_resid,1-alpha))
    
    for i,key in enumerate(Keys_a):
        c = key//ka**2
        b = (key-(c*(ka**2)))//ka
        a = key%ka

        dict_residuals[(a,b,c)]=array_resid[3*i+0]+array_resid[3*i+1]+array_resid[3*i+2]
        
    return(dict_residuals)
